package attractor {

/**
 *	Class description.
 *
 *	@langversion ActionScript 3.0
 *	@playerversion Flash 9.0
 *
 *	@author Zach Archer
 *	@since  09.08.2009
 */

import flash.display.*;
import flash.events.*;
import flash.geom.*;
import attractor.*;

public class Main extends Sprite
{
	//--------------------------------------
	// CLASS CONSTANTS
	//--------------------------------------
	
	//--------------------------------------
	//  CONSTRUCTOR
	//--------------------------------------
	public function Main() {
		trace("Hello world!");
		
		stage.scaleMode = StageScaleMode.NO_SCALE;
		stage.align = StageAlign.TOP_LEFT;
		
		_bData = new BitmapData( 800, 800, false, 0x0 );
		_bmp = new Bitmap( _bData, PixelSnapping.ALWAYS, false );
		addChildAt( _bmp, 0 );
		
		stage.addEventListener( MouseEvent.CLICK, clickHandler );
		
		_recreateNow = true;
		addEventListener( Event.ENTER_FRAME, enterFrameHandler );
		
		tf.htmlText = '<a href="http://blog.zacharcher.com/">blog.zacharcher.com</a>';

		stage.addEventListener( Event.RESIZE, resizeHandler );
		resizeHandler();
	}
	
	//--------------------------------------
	//  PRIVATE VARIABLES
	//--------------------------------------
	private var _bData :BitmapData;
	private var _bmp :Bitmap;
	private var _userClicked :Boolean = false;
	private var _recreateNow :Boolean = false;
	
	// Attractor state
	private var _X :Number = 1;
	private var _Y :Number = 1;
	private var _showRect :Rectangle;
	
	// Attractor coefficients
	private var ax :Number;
	private var ax_x :Number;
	private var ax_xx :Number;
	private var ax_xy :Number;
	private var ax_y :Number;
	private var ax_yy :Number;
	private var ay :Number;
	private var ay_x :Number;
	private var ay_xx :Number;
	private var ay_xy :Number;
	private var ay_y :Number;
	private var ay_yy :Number;
	
	//--------------------------------------
	//  GETTER/SETTERS
	//--------------------------------------
	
	//--------------------------------------
	//  PUBLIC METHODS
	//--------------------------------------
	
	//--------------------------------------
	//  EVENT HANDLERS
	//--------------------------------------
	private function clickHandler( e:MouseEvent ) :void {
		_bData.fillRect( _bData.rect, 0x0 );
		_userClicked = true;
	}
	
	private function enterFrameHandler( e:Event ) :void {
		if( _userClicked ) {
			// Finding a nice attractor takes a while, so...
			// Immediately blank the image, before creating a new attractor.
			_userClicked = false;
			_recreateNow = true;
			_bData.fillRect( _bData.rect, 0x0 );
			return;
			
		} else if( _recreateNow ) {
			_recreateNow = false;
			createNewAttractor();
			return;
		}
		
		for( var i:int=0; i<20000; i++ ) {
			var newX:Number = ax + ax_x*_X + ax_xx*(_X*_X) + ax_xy*(_X*_Y) + ax_y*(_Y) + ax_yy*(_Y*_Y);
			var newY:Number = ay + ay_x*_X + ay_xx*(_X*_X) + ay_xy*(_X*_Y) + ay_y*(_Y) + ay_yy*(_Y*_Y);		
			_X = newX;
			_Y = newY;
			
			var plotX:int = ((_X - _showRect.left) / _showRect.width) * 800.0;
			var plotY:int = ((_Y - _showRect.top) / _showRect.height) * 800.0;
			
			//trace("wat", plot.x, plot.y);
			var rgb:uint = _bData.getPixel( plotX, plotY );
			if( (rgb & 0x00ff00) < 0xff00 ) {	// fill in green first
				_bData.setPixel( plotX, plotY, rgb+0x0100 );
			} else if( (rgb & 0x0000ff) < 0xff ) {	// then blue
				_bData.setPixel( plotX, plotY, rgb+0x01 );
			} else if( rgb < 0xff0000 ) {	// finally white
				_bData.setPixel( plotX, plotY, rgb+0x010000 );
			}
		}
	}
	
	private function resizeHandler( e:Event=null ) :void {
		tf.x = 5;
		tf.y = stage.stageHeight - 21;
		
		_bmp.x = (stage.stageWidth-800) / 2;
		_bmp.y = (stage.stageHeight-820) / 2;
	}
	
	//--------------------------------------
	//  PRIVATE & PROTECTED INSTANCE METHODS
	//--------------------------------------
	private function createNewAttractor() {
		// We need to choose coefficients that are stable.
		// Generate new coefficients, run some number of iterations -- does the rectangle stabilize to a reasonable degree?
		var isStable:Boolean = false;
		var creationAttempts:int = 0;

		var prefix:String, suffix:String;
		
		while( !isStable ) {
			// Choose coefficients
			for each( prefix in "ax,ay".split(',') ) {
				for each( suffix in ",_x,_xx,_xy,_y,_yy".split(',') ) {
					// Bipolar random numbers
					this[prefix+suffix] = (Math.random() * 2.0) - 1.0;
					//trace( prefix+suffix, ":", this[prefix+suffix] );
				}
			}
			_X = 0;
			_Y = 0;
			
			// Discard the first N attractor advances, give it a chance to stabilize.
			var newX:Number, newY:Number;
			_bData.fillRect( _bData.rect, 0x0 );
			var maxBright:uint = 0x0;
			
			var i:int;
			for( i=0; i<5000; i++ ) {
				newX = ax + ax_x*_X + ax_xx*(_X*_X) + ax_xy*(_X*_Y) + ax_y*(_Y) + ax_yy*(_Y*_Y);
				newY = ay + ay_x*_X + ay_xx*(_X*_X) + ay_xy*(_X*_Y) + ay_y*(_Y) + ay_yy*(_Y*_Y);		
				_X = newX;
				_Y = newY;
			}
			
			// Advance the attractor some number of steps, then advance another set of steps.
			// If the bounding rects are similar sizes, then the attractor is probably stable.
			// Try to avoid single loops; discard coeffs that repeat pixels too often.
			var rects:Vector.<Rectangle> = new Vector.<Rectangle>();
			
			for( var r:int=0; r<2; r++ ) {
				var rect:Rectangle = new Rectangle( _X, _Y, 0, 0 );
				for( i=0; i<5000; i++ ) {
					newX = ax + ax_x*_X + ax_xx*(_X*_X) + ax_xy*(_X*_Y) + ax_y*(_Y) + ax_yy*(_Y*_Y);
					newY = ay + ay_x*_X + ay_xx*(_X*_X) + ay_xy*(_X*_Y) + ay_y*(_Y) + ay_yy*(_Y*_Y);		
					_X = newX;
					_Y = newY;
					
					if( _X < rect.left ) {
						rect.left = _X;
					} else if( _X > rect.right ) {
						rect.right = _X;
					}
					if( _Y < rect.top ) {
						rect.top = _Y;
					} else if( _Y > rect.bottom ) {
						rect.bottom = _Y;
					}
				}
				rects.push( rect );
			}
			
			// Are these two rectangles fairly similar? Compare their sizes:
			var r0:Rectangle = rects[0];
			var r1:Rectangle = rects[1];
			var ratioX:Number = r0.width / r1.width;
			if( ratioX < 1 ) ratioX = 1/ratioX;	// always >= 1.0
			var ratioY:Number = r0.height / r1.height;
			if( ratioY < 1 ) ratioY = 1/ratioY;	// always >= 1.0
			// However, some collapsing systems have ratios that are extremely near 1: these are bad, throw them out
			if( Math.max(ratioX,ratioY) < 1.1 && Math.min(ratioX,ratioY) > 1.00000001 ) {
				
				// Using the last rectangle, run 5000 more steps and plot the results.
				// We want to avoid boring shapes (i.e. simple loops) -- if these coefficients have too many repeated pixels,
				// then this is not a success.
				_bData.fillRect( _bData.rect, 0x0 );
				_showRect = r1;
				var replotCount:int = 0;
				for( i=0; i<5000; i++ ) {
					newX = ax + ax_x*_X + ax_xx*(_X*_X) + ax_xy*(_X*_Y) + ax_y*(_Y) + ax_yy*(_Y*_Y);
					newY = ay + ay_x*_X + ay_xx*(_X*_X) + ay_xy*(_X*_Y) + ay_y*(_Y) + ay_yy*(_Y*_Y);		
					_X = newX;
					_Y = newY;

					var plotX:int = ((_X - _showRect.left) / _showRect.width) * 800.0;
					var plotY:int = ((_Y - _showRect.top) / _showRect.height) * 800.0;
					var rgb:uint = _bData.getPixel( plotX, plotY );
					if( rgb > 0 ) replotCount++;
					_bData.setPixel( plotX, plotY, rgb+1 );
				}
				
				var replotRatio:Number = Number(replotCount) / 5000.0;
				//trace("REPLOT RATIO:", replotRatio);
				// 800x800, 5000 iterations: anything above 0.2 is fairly uninteresting
				// For some reason, it is possible for blank attractors with replotRatio 0 to sneak through, so prevent those too
				if( 0 < replotRatio && replotRatio < 0.15 ) {
					trace("-------------------------------------------------------");
					// Less than ~1.1 size change, so it's probably stable...?
					isStable = true;
					trace("here are rect ratios:", ratioX, ratioY);
					trace("here are rects:", r0, r1 );
					trace("replot ratio:", replotRatio);
				}
			}
			
			creationAttempts++;
		}
		
		trace("Found a stable attractor after", creationAttempts, "attempts");
		var coeffs:Array = [];
		for each( prefix in "ax,ay".split(',') ) {
			for each( suffix in ",_x,_xx,_xy,_y,_yy".split(',') ) {
				coeffs.push( prefix+suffix+'='+this[prefix+suffix]);
			}
		}
		trace( coeffs.join(',') );
		
		_bData.fillRect( _bData.rect, 0x0 );
	}
	
}

}

